#!/usr/bin/env python3

import re
import sys
import unittest


SIMPLE_ESCAPE = {
    ord("'"): ord("'"),
    ord('"'): ord('"'),
    ord('?'): ord('?'),
    ord('\\'): ord('\\'),
    ord('a'): b'\a'[0],
    ord('b'): b'\b'[0],
    ord('f'): b'\f'[0],
    ord('n'): b'\n'[0],
    ord('r'): b'\r'[0],
    ord('t'): b'\t'[0],
    ord('v'): b'\v'[0],
}


RE_HEX = re.compile(br'\\u[0-9a-fA-F]{4}')
RE_UNICODE_LONG = re.compile(br'\\U[0-9a-fA-F]{8}')
RE_UNICODE_SHORT = re.compile(br'\\x[0-9a-fA-F]{2}')
RE_OCT = re.compile(br'\\[0-9]{1,3}')


class LexerFoundDblQuoteException(Exception):
    pass


class LexerInvalidStartOrEndException(Exception):
    pass


def noescape(xs, strict=True, verbose=False):

    if (xs[0] != b'"'[0]) or (xs[-1] != b'"'[0]):
        raise LexerInvalidStartOrEndException('invalid start/end')

    xs = xs[1:-1]

    ret = bytearray(len(xs) + 10)
    ret_len = 0
    xs_ptr = 0
    xs_len = len(xs)

    if not xs_len:
        return bytes(ret[:0])

    while xs_ptr != xs_len:
        if verbose:
            print(xs[xs_ptr:])

        if len(ret) - ret_len < 10:
            ret += bytearray(10)

        while xs_ptr != xs_len and xs[xs_ptr] not in br'\"':
            ret[ret_len] = xs[xs_ptr]
            xs_ptr += 1
            ret_len += 1

        if xs_ptr == xs_len:
            break

        if strict:
            if xs[xs_ptr] == b'"'[0]:
                raise LexerFoundDblQuoteException('found unescaped " in stream')

        if xs[xs_ptr] == b'\\'[0]:
            if xs_ptr + 1 == xs_len:
                ret[ret_len] = xs[xs_ptr]
                xs_ptr += 1
                ret_len += 1
                break

            # \n, \", etc...
            ch = xs[xs_ptr + 1]
            if ch in SIMPLE_ESCAPE:
                ret[ret_len] = SIMPLE_ESCAPE[ch]
                ret_len += 1
                xs_ptr += 2
                continue
            elif ch == b'u'[0]:
                try:
                    m = RE_HEX.match(xs[xs_ptr:xs_ptr+4+2])
                    if m:
                        gg = m.group(0)
                        bb = gg.decode('unicode_escape').encode('utf8')
                        for b in bb:
                            ret[ret_len] = b
                            ret_len += 1
                        xs_ptr += len(gg)
                        continue
                except UnicodeDecodeError as e:
                    if verbose:
                        print(e)
                except UnicodeEncodeError as e:
                    if verbose:
                        print(e)
            elif ch == b'U'[0]:
                try:
                    m = RE_UNICODE_LONG.match(xs[xs_ptr:xs_ptr+8+2])
                    if m:
                        gg = m.group(0)
                        bb = gg.decode('unicode_escape').encode('utf8')
                        for b in bb:
                            ret[ret_len] = b
                            ret_len += 1
                        xs_ptr += len(gg)
                        continue
                except UnicodeDecodeError as e:
                    if verbose:
                        print(e)
            elif ch == b'x'[0]:
                try:
                    m = RE_UNICODE_SHORT.match(xs[xs_ptr:xs_ptr+2+2])
                    if m:
                        gg = m.group(0)
                        val = int(gg[2:], 16)
                        ret[ret_len] = val
                        ret_len += 1
                        xs_ptr += len(gg)
                        continue
                except IndexError as e:
                    if verbose:
                        print(e)
            elif ch in b'0123456789':
                try:
                    m = RE_OCT.match(xs[xs_ptr:xs_ptr+3+1])
                    if m:
                        gg = m.group(0)
                        val = int(gg[1:], 8)
                        ret[ret_len] = val
                        ret_len += 1
                        xs_ptr += len(gg)
                        continue
                except ValueError as e:
                    if verbose:
                        print(e)
            if xs_ptr == xs_len:
                break
            ret[ret_len] = xs[xs_ptr]
            xs_ptr += 1
            ret_len += 1

    return bytes(ret[:ret_len])


class TestNoEscape(unittest.TestCase):

    def loop_equal(self, tests):
        for xxx, yyy in tests:
            self.assertEqual(noescape(b'"' + xxx + b'"'), yyy)

    def test_general(self):
        tests = [
            (br'',              b''),
            (br'o',             b'o'),
            (br'\"\n\a\f\b\r',  b'"\n\a\f\b\r'),
        ]
        self.loop_equal(tests)

    def test_random(self):
        import random
        
        for i in range(100):
            test = 'b\'"'
            for n in range(random.randint(0, 50)):
                test += random.choice(
                        '0123456789abcdefABCDEFXuU\\'
                        )
            test += '"\''
            try:
                unescaped = eval(test)
                # print(unescaped, test)
                self.assertEqual(noescape(test), unescaped)  # YOLO
            except SyntaxError:
                continue
            except LexerInvalidStartOrEndException:
                continue

    def test_hex(self):
        tests = [
            (br'\x',        br'\x'),
            (br'\x00',      b'\0'),
            (br'\x0',       br'\x0'),
            (br'\xFF',      b'\xFF'),
            (br'\x\x\xFF',  b'\\x\\x\xFF'),
        ]
        self.loop_equal(tests)

    def test_oct(self):
        tests = [
            (br'\0',        b'\0'),
            (br'\00',       b'\0'),
            (br'\000',      b'\0'),
            (br'\0000',     b'\0000'),
            (br'0\0000',    b'0\0000'),
            (br'\60',       b'0'),
        ]
        self.loop_equal(tests)

    def test_unicode(self):
        tests = [
            (br'\u0000',        '\u0000'.encode('utf8')),
            (br'\uFFFF',        '\uFFFF'.encode('utf8')),
            (br'\u0030',        b'\x30'),
            (br'\u003',         br'\u003'),
            (br'\u003xxx',      br'\u003xxx'),
            (br'xxx\u003',      br'xxx\u003'),
            (br'\U00000000',    b'\0'),
            (br'\U10000000',    br'\U10000000'),  # invalid utf code
            (br'\U00010027',    chr(0x00010027).encode('utf8')),  # invalid utf code
            (br'[\"\\ude36\"]', b'[\"\\ude36\"]'),
        ]
        self.loop_equal(tests)


